@@ -28,6 +28,7 @@
 #include <unistd.h>
 #include <string.h>
 #include <signal.h>
+#include <sys/signalfd.h>
 
 #include <gdbus.h>
 
@@ -39,18 +40,74 @@
 
 static GMainLoop *main_loop = NULL;
 
-static volatile sig_atomic_t __terminated = 0;
+static unsigned int __terminated = 0;
 
-static void sig_term(int sig)
+static gboolean signal_handler(GIOChannel *channel, GIOCondition cond,
+							gpointer user_data)
 {
-	if (__terminated > 0)
-		return;
+	struct signalfd_siginfo si;
+	ssize_t result;
+	int fd;
 
-	__terminated = 1;
+	if (cond & (G_IO_NVAL | G_IO_ERR | G_IO_HUP))
+		return FALSE;
 
-	pacrunner_info("Terminating");
+	fd = g_io_channel_unix_get_fd(channel);
 
-	g_main_loop_quit(main_loop);
+	result = read(fd, &si, sizeof(si));
+	if (result != sizeof(si))
+		return FALSE;
+
+	switch (si.ssi_signo) {
+	case SIGINT:
+	case SIGTERM:
+		if (__terminated == 0) {
+			pacrunner_info("Terminating");
+			g_main_loop_quit(main_loop);
+		}
+
+		__terminated = 1;
+		break;
+	}
+
+	return TRUE;
+}
+
+static guint setup_signalfd(void)
+{
+	GIOChannel *channel;
+	guint source;
+	sigset_t mask;
+	int fd;
+
+	sigemptyset(&mask);
+	sigaddset(&mask, SIGINT);
+	sigaddset(&mask, SIGTERM);
+
+	if (sigprocmask(SIG_BLOCK, &mask, NULL) < 0) {
+		perror("Failed to set signal mask");
+		return 0;
+	}
+
+	fd = signalfd(-1, &mask, 0);
+	if (fd < 0) {
+		perror("Failed to create signal descriptor");
+		return 0;
+	}
+
+	channel = g_io_channel_unix_new(fd);
+
+	g_io_channel_set_close_on_unref(channel, TRUE);
+	g_io_channel_set_encoding(channel, NULL, NULL);
+	g_io_channel_set_buffered(channel, FALSE);
+
+	source = g_io_add_watch(channel,
+				G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
+				signal_handler, NULL);
+
+	g_io_channel_unref(channel);
+
+	return source;
 }
 
 static void disconnect_callback(DBusConnection *conn, void *user_data)
@@ -99,7 +156,7 @@ int main(int argc, char *argv[])
 	GError *error = NULL;
 	DBusConnection *conn;
 	DBusError err;
-	struct sigaction sa;
+	guint signal;
 
 #ifdef HAVE_CAPNG
 	/* Drop capabilities */
@@ -143,6 +200,8 @@ int main(int argc, char *argv[])
 		exit(1);
 	}
 
+	signal = setup_signalfd();
+
 	dbus_error_init(&err);
 
 	conn = g_dbus_setup_bus(DBUS_BUS_SYSTEM, PACRUNNER_SERVICE, &err);
@@ -170,13 +229,10 @@ int main(int argc, char *argv[])
 	g_free(option_plugin);
 	g_free(option_noplugin);
 
-	memset(&sa, 0, sizeof(sa));
-	sa.sa_handler = sig_term;
-	sigaction(SIGINT, &sa, NULL);
-	sigaction(SIGTERM, &sa, NULL);
-
 	g_main_loop_run(main_loop);
 
+	g_source_remove(signal);
+
 	__pacrunner_plugin_cleanup();
 	__pacrunner_manual_cleanup();
 	__pacrunner_client_cleanup();
